﻿
using System;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.DataTypeModel;
using OpcLabs.BaseLibFactory.DataTypeModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.BaseLibTest.DataTypeModel
{
    public partial class SequenceDataTypeTest
    {
        #region Static Private Properties

        [NotNull]
        static private SequenceDataType Circular1
        {
            get
            {
                var circular1 = new SequenceDataType(new PrimitiveDataType(typeof(int)));
                circular1.ElementDataType = circular1;
                return circular1;
            }
        }

        [NotNull]
        static private SequenceDataType Circular2
        {
            get
            {
                var intermediary = new StructuredDataType("Intermediary");

                var circular2 = new SequenceDataType(intermediary);

                intermediary.DataFields.Add(new DataField("idf1", new PrimitiveDataType(typeof(int))));
                intermediary.DataFields.Add(new DataField("idf2", circular2));
                intermediary.DataFields.Add(new DataField("idf3", new PrimitiveDataType(typeof(int))));

                return circular2;
            }
        }

        [NotNull]
        static private SequenceDataType Interminable1
        {
            get
            {
                var circular1 = new SequenceDataType(2, null);
                circular1.ElementDataType = circular1;
                return circular1;
            }
        }

        [NotNull]
        static private SequenceDataType OfPrimitive => new SequenceDataType(new PrimitiveDataType(typeof(int)));

        #endregion



        #region Test Methods

        [TestMethod]
        public void EqualsTest_Circular1_Circular2()
        {
            bool b = Circular1.Equals(Circular2);
            Assert.IsFalse(b);
        }

        [ExpectedException(typeof(ArgumentException))]
        [TestMethod]
        public void EqualsTest_Circular1_DifferentInstances()
        {
            // ReSharper disable once EqualExpressionComparison
            // ReSharper disable once ReturnValueOfPureMethodIsNotUsed
            Circular1.Equals(Circular1);
        }

        [TestMethod]
        public void EqualsTest_Circular1_SameInstances()
        {
            SequenceDataType instance = Circular1;
            // ReSharper disable once EqualExpressionComparison
            bool b = instance.Equals(instance);
            Assert.IsTrue(b);
        }

        [TestMethod]
        public void EqualsTest_Circular2_Circular1()
        {
            bool b = Circular2.Equals(Circular1);
            Assert.IsFalse(b);
        }

        [ExpectedException(typeof(ArgumentException))]
        [TestMethod]
        public void EqualsTest_Circular2_DifferentInstances()
        {
            // ReSharper disable once EqualExpressionComparison
            // ReSharper disable once ReturnValueOfPureMethodIsNotUsed
            Circular2.Equals(Circular2);
        }

        [TestMethod]
        public void EqualsTest_Circular2_SameInstances()
        {
            SequenceDataType instance = Circular2;
            // ReSharper disable once EqualExpressionComparison
            bool b = instance.Equals(instance);
            Assert.IsTrue(b);
        }

        [TestMethod]
        public void GetHashCodeTest_Circular1()
        {
            // ReSharper disable once ReturnValueOfPureMethodIsNotUsed
            Circular1.GetHashCode();
            // must not cause exception due to cycles
        }

        [TestMethod]
        public void GetHashCodeTest_Circular2()
        {
            // ReSharper disable once ReturnValueOfPureMethodIsNotUsed
            Circular2.GetHashCode();
            // must not cause exception due to cycles
        }

        [TestMethod]
        public void IsCompleteTest_Negatives()
        {
            Assert.IsFalse(SequenceDataTypeInstances.IncompleteInstance1.IsComplete);
            Assert.IsFalse(SequenceDataTypeInstances.IncompleteInstance2.IsComplete);
        }

        [TestMethod]
        public void IsCompleteTest_Positives()
        {
            Assert.IsTrue(SequenceDataTypeInstances.Instance0.IsComplete);
            Assert.IsTrue(SequenceDataTypeInstances.Instance1.IsComplete);
            Assert.IsTrue(SequenceDataTypeInstances.Instance2.IsComplete);
            Assert.IsTrue(SequenceDataTypeInstances.Instance3.IsComplete);
            Assert.IsTrue(SequenceDataTypeInstances.Instance4.IsComplete);
            Assert.IsTrue(SequenceDataTypeInstances.Instance5.IsComplete);
        }

        [TestMethod]
        public void IsRecursiveTest_Circular1()
        {
            Assert.IsTrue(Circular1.IsRecursive);
        }

        [TestMethod]
        public void IsRecursiveTest_Circular2()
        {
            Assert.IsTrue(Circular2.IsRecursive);
        }

        [TestMethod]
        public void IsRecursiveTest_OfPrimitive()
        {
            Assert.IsFalse(OfPrimitive.IsRecursive);
        }

        [TestMethod]
        public void IsTerminableTest_Circular1()
        {
            Assert.IsTrue(Circular1.IsTerminable);
        }

        [TestMethod]
        public void IsTerminableTest_Interminable1()
        {
            Assert.IsFalse(Interminable1.IsTerminable);
        }

        [TestMethod]
        public void IsTerminableTest_OfPrimitive()
        {
            Assert.IsTrue(OfPrimitive.IsTerminable);
        }

        [TestMethod]
        public void ToStringTest_Circular1()
        {
            SequenceDataType target = Circular1;
            string s = target.ToString();
            Assert.IsNotNull(s);
            AssertUtilities.WriteLine(s);
            Assert.AreEqual("sequence[*] of ", s);
        }

        [TestMethod]
        public void ToStringTest_Circular1_Verbose()
        {
            SequenceDataType target = Circular1;
            string s = target.ToString("V", null);
            AssertUtilities.WriteLine(s);
            Assert.IsNotNull(s);
            // IMPROVE: check the output
        }

        [TestMethod]
        public void ToStringTest_Circular2()
        {
            SequenceDataType target = Circular2;
            string s = target.ToString();
            Assert.IsNotNull(s);
            AssertUtilities.WriteLine(s);
            Assert.AreEqual("sequence[*] of Intermediary", s);
        }

        [TestMethod]
        public void ToStringTest_Circular2_Verbose()
        {
            SequenceDataType target = Circular2;
            string s = target.ToString("V", null);
            AssertUtilities.WriteLine(s);
            Assert.IsNotNull(s);
            // IMPROVE: check the output
        }

        #endregion
    }
}
