﻿
using System;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Extensions.Internal;
using OpcLabs.BaseLib.Runtime.InteropServices;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.AlarmsAndEvents;
using OpcLabs.EasyOpc.AlarmsAndEvents.AddressSpace;
using OpcLabs.EasyOpc.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.AlarmsAndEvents
{
    partial class EasyAEClientTest
    {
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseNodesTest_Basic_Root()
        {
            BrowseNodesTest_Basic(AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId, AENodeDescriptor.Root, 1);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseNodesTest_Basic_UnknownNodeId()
        {
            BrowseNodesTest_UnknownNodeId(AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId);
        }



        private void BrowseNodesTest_Basic(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] AENodeDescriptor parentNodeDescriptor,
            int expectedCount)  // -1 for "unknown"
        {
            TestSection.Arrange();
            TestEnvironment.PrecheckServer(serverDescriptor);
            var client = new EasyAEClient();

            TestSection.Act();
            AENodeElementCollection nodeElements = client.BrowseNodes(
                serverDescriptor,
                parentNodeDescriptor,
                new AEBrowseParameters(AEBrowseFilter.Areas));

            TestSection.Assert();
            Assert.IsNotNull(nodeElements);
            AssertUtilities.WriteLine(nodeElements.Count);
            CollectionAssert.AllItemsAreNotNull(nodeElements);
            foreach (AENodeElement nodeElement in nodeElements)
                AssertUtilities.WriteLine($"nodeElement: {nodeElement}");
            if (expectedCount >= 0)
                Assert.AreEqual(expectedCount, nodeElements.Count);
        }

        protected void BrowseNodesTest_UnknownNodeId([NotNull] ServerDescriptor serverDescriptor)
        {
            TestSection.Arrange();
            TestEnvironment.PrecheckServer(serverDescriptor);
            var client = new EasyAEClient();

            AENodeElementCollection actualResultCollection;
            try
            {
                TestSection.Act();
                try
                {
                    actualResultCollection = client.BrowseAreas(serverDescriptor, "unknownNodeId");
                }
                finally
                {
                    TestSection.Assert();
                }
            }
            catch (OpcException opcException)
            {
                Exception innerException = opcException.InnerException;
                Assert.IsNotNull(innerException);

                Exception baseException = innerException.GetBaseException();
                Assert.IsNotNull(baseException);
                AssertUtilities.WriteLine($"baseException: {baseException}");
                Assert.IsInstanceOfType(baseException, typeof(ExtendedCOMException));
                var extendedComException = (ExtendedCOMException)baseException;
                AssertUtilities.WriteLine($"extendedComException.QualifiedName: {extendedComException.QualifiedName}");
                AssertUtilities.WriteLine($"extendedComException.GetErrorId(): {extendedComException.GetErrorId()}");
                // ReSharper disable once CommentTypo
                Assert.AreEqual(unchecked((int)0xC0040203), extendedComException.ErrorCode);   // OPC_E_INVALIDBRANCHNAME
                Assert.AreEqual("http://opcfoundation.org/AlarmAndEvents/", extendedComException.QualifiedName.Namespace);
                // ReSharper disable once StringLiteralTypo
                Assert.AreEqual("E_INVALIDBRANCHNAME", extendedComException.QualifiedName.Name);

                return;
            }
            AssertUtilities.WriteLine(actualResultCollection.Count);
            Assert.Fail();
        }
    }
}
