﻿using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.Extensions;
using OpcLabs.EasyOpcLiveTest.DataAccess.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.DataAccess.Extensions
{
    [TestClass]
    public class DAItemPropertyRecordTest
    {

#if (NETFRAMEWORK)
        [TestCategory("BinarySerialization")]
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BinarySerializationTestLive_Com_Native()
        {
            Tests.BinarySerializationTest(CreateSerializationTargetLive_Com(new EasyDAClient {InstanceParameters = {EnableNetApiClient = false}}));
        }
#endif

#if (NETFRAMEWORK)
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("XmlSerialization")]
        [TestMethod]
        public void XmlSerializationTestLive_Com_Native()
        {
            Tests.XmlSerializationTest(CreateSerializationTargetLive_Com(new EasyDAClient {InstanceParameters = { EnableNetApiClient = false }}));
        }
#endif

        [TestCategory("BinarySerialization")]
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BinarySerializationTestLive_Com_NetApi()
        {
            Tests.BinarySerializationTest(CreateSerializationTargetLive_Com(new EasyDAClient { InstanceParameters = { EnableNativeClient = false } }));
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("XmlSerialization")]
        [TestMethod]
        public void XmlSerializationTestLive_Com_NetApi()
        {
            Tests.XmlSerializationTest(CreateSerializationTargetLive_Com(new EasyDAClient { InstanceParameters = { EnableNativeClient = false } }));
        }

        [TestCategory("BinarySerialization")]
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BinarySerializationTestLive_Xml()
        {
            Tests.BinarySerializationTest(CreateSerializationTargetLive_Xml(new EasyDAClient()));
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("XmlSerialization")]
        [TestMethod]
        public void XmlSerializationTestLive_Xml()
        {
            Tests.XmlSerializationTest(CreateSerializationTargetLive_Xml(new EasyDAClient()));
        }



        [NotNull]
        private DAItemPropertyRecord CreateSerializationTargetLive_Com([NotNull] IEasyDAClient client)
        {
            DAItemPropertyRecord itemPropertyRecord = client.GetItemPropertyRecord(
                new ServerDescriptor("", OpcLabs_KitServer.ProgId), 
                OpcLabs_KitServer.ItemIds.Random/*,
                DAItemPropertyRecord.PropertyIds,
                allowFailures: false*/);
            AssertUtilities.WriteLine($"itemPropertyRecord: {itemPropertyRecord}");
            foreach (DAPropertyId propertyId in DAItemPropertyRecord.PropertyIds)
                // ReSharper disable once AssignNullToNotNullAttribute
                AssertUtilities.WriteLine($"  {propertyId}: {itemPropertyRecord.GetProperty(propertyId)}");
            return itemPropertyRecord;
        }

        [NotNull]
        private DAItemPropertyRecord CreateSerializationTargetLive_Xml([NotNull] IEasyDAClient client)
        {
            return client.GetItemPropertyRecord(DALiveTestEnvironment.OpcFoundation_XmlDaSampleServer.ServerDescriptors.Http,
                "Dynamic/Analog Types/Fools");
        }
    }
}
