﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Threading;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.ComInterop;
using OpcLabs.BaseLib.Net.Sockets;
using OpcLabs.BaseLib.Runtime;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.AddressSpace;
using OpcLabs.EasyOpc.DataAccess.Extensions;
using OpcLabs.EasyOpc.DataAccess.OperationModel;
using OpcLabs.EasyOpc.OperationModel;
using OpcLabs.EasyOpcLiveTest.DataAccess.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.DataAccess
{
    // ReSharper disable once InconsistentNaming
    public class IEasyDAClientCatalogTest_Functional : IEasyDAClientCatalogTestBase
    {
        // IMPROVE: have LocationCatalog, and find locations there

        #region Public Constructors

        public IEasyDAClientCatalogTest_Functional(
            [NotNull] Func<IEasyDAClient> createClientFunc,
            [NotNull] DAServerCatalog serverCatalog)
            : base(createClientFunc, serverCatalog)
        {
        }

        #endregion



        #region Test Methods

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseNodesTest_Basic_Dcom_Root()
        {
            IEnumerable<DAServerCatalogEntry> serverCatalogEntries =
                // ReSharper disable once PossibleNullReferenceException
                ServerCatalog.Where(entry => entry.Element.Technology == OpcTechnologies.Com);

            foreach (DAServerCatalogEntry serverCatalogEntry in serverCatalogEntries)
            {
                Debug.Assert(serverCatalogEntry != null);
                ServerDescriptor serverDescriptor = serverCatalogEntry.Element;
                // ReSharper disable once PossibleNullReferenceException
                serverDescriptor.Scheme = "dcom";
                AssertUtilities.WriteLine("");
                AssertUtilities.WriteLine(serverDescriptor);
                BrowseNodesTest_Basic(serverDescriptor, DANodeDescriptor.Root);
            }
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseNodesTest_Basic_Http_Root()
        {
            IEnumerable<DAServerCatalogEntry> serverCatalogEntries =
                // ReSharper disable once PossibleNullReferenceException
                ServerCatalog.Where(entry => entry.Element.Technology == OpcTechnologies.Xml);

            bool inconclusive = false;
            foreach (DAServerCatalogEntry serverCatalogEntry in serverCatalogEntries)
            {
                Debug.Assert(serverCatalogEntry != null);
                ServerDescriptor serverDescriptor = serverCatalogEntry.Element;
                // ReSharper disable once PossibleNullReferenceException
                serverDescriptor.Scheme = "http";
                AssertUtilities.WriteLine(serverDescriptor);
                if (!TcpClientUtilities.CanCreate(serverDescriptor))
                {
                    inconclusive = true;
                    continue;
                }
                BrowseNodesTest_Basic(serverDescriptor, DANodeDescriptor.Root);
            }
            if (inconclusive)
                Assert.Inconclusive("Failed pre-check: Socket exception.");
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        // ReSharper disable once IdentifierTypo
        public void BrowseNodesTest_Basic_Opcda_Root()
        {
            IEnumerable<DAServerCatalogEntry> serverCatalogEntries =
                // ReSharper disable once PossibleNullReferenceException
                ServerCatalog.Where(entry => entry.Element.Technology == OpcTechnologies.Com);

            foreach (DAServerCatalogEntry serverCatalogEntry in serverCatalogEntries)
            {
                Debug.Assert(serverCatalogEntry != null);
                ServerDescriptor serverDescriptor = serverCatalogEntry.Element;
                // ReSharper disable once PossibleNullReferenceException
                // ReSharper disable once StringLiteralTypo
                serverDescriptor.Scheme = "opcda";
                AssertUtilities.WriteLine(serverDescriptor);
                BrowseNodesTest_Basic(serverDescriptor, DANodeDescriptor.Root);
            }
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void GetPropertyValueCatalogTests_Basic()
        {
            var tests = new List<Action>();
            foreach (DAServerCatalogEntry serverCatalogEntry in ServerCatalog)
            {
                Debug.Assert(serverCatalogEntry != null);

                IEnumerable<DAItemCatalogEntry> itemCatalogEntries = serverCatalogEntry.GetReadableItems();
                foreach (DAItemCatalogEntry itemCatalogEntry in itemCatalogEntries)
                {
                    Debug.Assert(itemCatalogEntry != null);

                    foreach (DAPropertyCatalogEntry propertyCatalogEntry in itemCatalogEntry.PropertyCatalog)
                    {
                        Debug.Assert(propertyCatalogEntry != null);

                        VarType dataType = propertyCatalogEntry.Element.DataType;
                        Type expectedType = null;
                        if (dataType != VarTypes.Empty)
                            expectedType = Cls.GetClsType(VarTypeUtilities.TypeFromVarType(dataType));

                        DAServerCatalogEntry server = serverCatalogEntry;
                        DAItemCatalogEntry item = itemCatalogEntry;
                        DAPropertyCatalogEntry property = propertyCatalogEntry;
                        tests.Add(() => GetPropertyValueTest_Basic(server.Element, item.Element, property.Element, expectedType));
                    }
                }
            }
            TestUtilities.RunTests(tests);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void GetPropertyValueCatalogTests_Timestamp()
        {
            var tests = new List<Action>();
            foreach (DAServerCatalogEntry serverCatalogEntry in ServerCatalog)
            {
                Debug.Assert(serverCatalogEntry != null);

                IEnumerable<DAItemCatalogEntry> itemCatalogEntries = serverCatalogEntry.GetReadableItems();
                foreach (DAItemCatalogEntry itemCatalogEntry in itemCatalogEntries)
                {
                    Debug.Assert(itemCatalogEntry != null);

                    foreach (DAPropertyCatalogEntry propertyCatalogEntry in itemCatalogEntry.PropertyCatalog
                        // ReSharper disable once PossibleNullReferenceException
                        .Where(entry => entry.Element.PropertyId.InternalValue == DAPropertyIds.Timestamp))
                    {
                        Debug.Assert(propertyCatalogEntry != null);

                        DAServerCatalogEntry server = serverCatalogEntry;
                        DAItemCatalogEntry item = itemCatalogEntry;
                        DAPropertyCatalogEntry property = propertyCatalogEntry;
                        tests.Add(() => GetPropertyValueTest_Timestamp(server.Element, item.Element, property.Element));
                    }
                }
            }
            TestUtilities.RunTests(tests);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemCatalogTests_Basic()
        {
            var tests = new List<Action>();
            foreach (DAServerCatalogEntry serverCatalogEntry in ServerCatalog)
            {
                Debug.Assert(serverCatalogEntry != null);

                IEnumerable<DAItemCatalogEntry> itemCatalogEntries = serverCatalogEntry.GetReadableItems();
                foreach (DAItemCatalogEntry itemCatalogEntry in itemCatalogEntries)
                {
                    Debug.Assert(itemCatalogEntry != null);

                    VarType dataType = itemCatalogEntry.DataType;
                    Type expectedType = null;
                    if (dataType != VarTypes.Empty)
                        expectedType = Cls.GetClsType(VarTypeUtilities.TypeFromVarType(dataType));

                    DAServerCatalogEntry server = serverCatalogEntry;
                    DAItemCatalogEntry item = itemCatalogEntry;
                    tests.Add(() => ReadItemTest_Basic(server.Element, item.Element, expectedType));
                }
            }
            TestUtilities.RunTests(tests);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemCatalogTests_Constant_Basic()
        {
            var tests = new List<Action>();
            foreach (DAServerCatalogEntry serverCatalogEntry in ServerCatalog)
            {
                Debug.Assert(serverCatalogEntry != null);

                IEnumerable<DAItemCatalogEntry> itemCatalogEntries = serverCatalogEntry.GetConstantItems();
                foreach (DAItemCatalogEntry itemCatalogEntry in itemCatalogEntries)
                {
                    Debug.Assert(itemCatalogEntry != null);
                    DAServerCatalogEntry server = serverCatalogEntry;
                    DAItemCatalogEntry item = itemCatalogEntry;
                    tests.Add(() => ReadItemTest_Constant_Basic(
                        server.Element,
                        new DAItemDescriptor(item.Element),
                        DAReadParameters.Default,
                        item.Quality,
                        item.Value));
                }
            }
            TestUtilities.RunTests(tests);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemCatalogTests_Constant_Basic_Device()
        {
            var tests = new List<Action>();
            foreach (DAServerCatalogEntry serverCatalogEntry in ServerCatalog)
            {
                Debug.Assert(serverCatalogEntry != null);

                IEnumerable<DAItemCatalogEntry> itemCatalogEntries = serverCatalogEntry.GetConstantItems();
                foreach (DAItemCatalogEntry itemCatalogEntry in itemCatalogEntries)
                {
                    Debug.Assert(itemCatalogEntry != null);
                    DAServerCatalogEntry server = serverCatalogEntry;
                    DAItemCatalogEntry item = itemCatalogEntry;
                    tests.Add(() => ReadItemTest_Constant_Basic(
                        server.Element,
                        new DAItemDescriptor(item.Element),
                        DAReadParameters.DeviceSource,
                        item.Quality,
                        item.Value));
                }
            }
            TestUtilities.RunTests(tests);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemCatalogTests_NonReadable()
        {
            var tests = new List<Action>();
            foreach (DAServerCatalogEntry serverCatalogEntry in ServerCatalog)
            {
                Debug.Assert(serverCatalogEntry != null);

                IEnumerable<DAItemCatalogEntry> itemCatalogEntries = serverCatalogEntry.GetNonReadableItems();
                foreach (DAItemCatalogEntry itemCatalogEntry in itemCatalogEntries)
                {
                    Debug.Assert(itemCatalogEntry != null);

                    DAServerCatalogEntry server = serverCatalogEntry;
                    DAItemCatalogEntry item = itemCatalogEntry;
                    tests.Add(() => ReadItemTest_NonReadable(server.Element, item.Element));
                }
            }
            TestUtilities.RunTests(tests);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadMultipleItemsCatalogTest_Basic()
        {
            var argumentsList = new List<DAReadItemArguments>();
            var serverCatalogEntryList = new List<DAServerCatalogEntry>();
            var itemCatalogEntryList = new List<DAItemCatalogEntry>();
            var serverDescriptorSet = new HashSet<ServerDescriptor>();
            foreach (DAServerCatalogEntry serverCatalogEntry in ServerCatalog)
            {
                Debug.Assert(serverCatalogEntry != null);

                serverDescriptorSet.Add(serverCatalogEntry.Element);
                IEnumerable<DAItemCatalogEntry> itemCatalogEntries = serverCatalogEntry.GetReadableItems();
                foreach (DAItemCatalogEntry itemCatalogEntry in itemCatalogEntries)
                {
                    Debug.Assert(itemCatalogEntry != null);

                    serverCatalogEntryList.Add(serverCatalogEntry);
                    itemCatalogEntryList.Add(itemCatalogEntry);

                    var arguments = new DAReadItemArguments(serverCatalogEntry.Element, itemCatalogEntry.Element);    
                    argumentsList.Add(arguments);
                }
            }

            foreach (ServerDescriptor serverDescriptor in serverDescriptorSet)
                // ReSharper disable once AssignNullToNotNullAttribute
                TestEnvironment.PrecheckServer(serverDescriptor);

            DAReadItemArguments[] argumentsArray = argumentsList.ToArray();
            DAItemCatalogEntry[] itemCatalogEntryArray = itemCatalogEntryList.ToArray();
            int count = argumentsArray.Length;

            IEasyDAClient client = CreateClient();
            var stopwatch = new Stopwatch();
            stopwatch.Start();
            DAVtqResult[] resultArray = client.ReadMultipleItems(argumentsArray);
            stopwatch.Stop();
            AssertUtilities.WriteLine(stopwatch.Elapsed);
            
            Assert.IsNotNull(resultArray);
            Assert.AreEqual(count, resultArray.Length);
            for (int i = 0; i < count; i++)
            {
                DAServerCatalogEntry serverCatalogEntry = serverCatalogEntryList[i];
                Debug.Assert(serverCatalogEntry != null);
                DAItemCatalogEntry itemCatalogEntry = itemCatalogEntryArray[i];
                Debug.Assert(itemCatalogEntry != null);

                AssertUtilities.WriteLine("");
                AssertUtilities.WriteLine(i);
                AssertUtilities.WriteLine(serverCatalogEntry.Element);
                AssertUtilities.WriteLine(itemCatalogEntry.Element);

                DAVtqResult result = resultArray[i];
                Assert.IsNotNull(result);
                AssertUtilities.WriteLine(result);

                if (!client.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsFalse(result.Succeeded);
                    Assert.IsInstanceOfType(result.Exception, typeof(NotSupportedException));
                    continue;
                }

                Assert.IsNull(result.Exception);
                Assert.AreEqual(null, result.ErrorId);

                DAVtq actualVtq = result.Vtq;
                Assert.IsNotNull(actualVtq);

                Type expectedType = null;

                VarType dataType = itemCatalogEntry.DataType;
                if (dataType != VarTypes.Empty)
                    expectedType = Cls.GetClsType(VarTypeUtilities.TypeFromVarType(dataType));
                if (expectedType != null)
                {
                    object actualValue = actualVtq.Value;
                    Assert.IsNotNull(actualValue);
                    Assert.AreEqual(expectedType, actualValue.GetType());
                }

                Assert.AreEqual(DateTimeKind.Utc, actualVtq.Timestamp.Kind);
                Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MinValue);
                Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MaxValue);
                Assert.IsTrue(actualVtq.Timestamp >= DateTime.FromFileTimeUtc(0));
                Assert.IsTrue(actualVtq.Timestamp <= DateTime.Now);
            }
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void SubscribeItemCatalogTests_Basic()
        {
            var tests = new List<Action>();
            foreach (DAServerCatalogEntry serverCatalogEntry in ServerCatalog)
            {
                Debug.Assert(serverCatalogEntry != null);

                IEnumerable<DAItemCatalogEntry> itemCatalogEntries = serverCatalogEntry.GetReadableItems();
                foreach (DAItemCatalogEntry itemCatalogEntry in itemCatalogEntries)
                {
                    Debug.Assert(itemCatalogEntry != null);

                    VarType dataType = itemCatalogEntry.DataType;
                    Type expectedType = null;
                    if (dataType != VarTypes.Empty)
                        expectedType = Cls.GetClsType(VarTypeUtilities.TypeFromVarType(dataType));

                    DAServerCatalogEntry server = serverCatalogEntry;
                    DAItemCatalogEntry item = itemCatalogEntry;
                    tests.Add(() => SubscribeItemTest_Basic(server.Element, item.Element, expectedType));
                }
            }
            TestUtilities.RunTests(tests);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        // ReSharper disable once IdentifierTypo
        public void SubscribeItemTest_Basic_Opcda_Local_EmptyAuthority()
        {
            SubscribeItemTest_Basic(
                // ReSharper disable once StringLiteralTypo
                "opcda:///OPCLabs.KitServer.2/{c8a12f17-1e03-401e-b53d-6c654dd576da}",
                "Simulation.Constant_BOOL",
                typeof(Boolean));
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        // ReSharper disable once IdentifierTypo
        public void SubscribeItemTest_Basic_Opcda_Local_NoAuthority()
        {
            SubscribeItemTest_Basic(
                // ReSharper disable once StringLiteralTypo
                "opcda:OPCLabs.KitServer.2/{c8a12f17-1e03-401e-b53d-6c654dd576da}",
                "Simulation.Constant_BOOL",
                typeof(Boolean));
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        // ReSharper disable once IdentifierTypo
        public void SubscribeItemTest_Error_Opcda_Local_UnknownItemId()
        {
            SubscribeItemTest_Error(
                // ReSharper disable once StringLiteralTypo
                "opcda:///OPCLabs.KitServer.2/{c8a12f17-1e03-401e-b53d-6c654dd576da}",
                "unknown_item");
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        // ReSharper disable once IdentifierTypo
        public void SubscribeItemTest_Basic_Opcda_Remote()
        {
            string machineName = DALiveTestDefinitions.Locations.OpcComTest;
            if (string.IsNullOrEmpty(machineName))
                Assert.Inconclusive();

            SubscribeItemTest_Basic(
                // ReSharper disable once StringLiteralTypo
                "opcda://" + machineName + "/OPCLabs.KitServer.2/{c8a12f17-1e03-401e-b53d-6c654dd576da}", 
                "Simulation.Constant_BOOL", 
                typeof(Boolean));
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        // ReSharper disable once IdentifierTypo
        public void SubscribeItemTest_Basic_Opcda_ThisComputer()
        {
            string thisComputer = System.Environment.MachineName;
            SubscribeItemTest_Basic(
                // ReSharper disable once StringLiteralTypo
                "opcda://" + thisComputer + "/OPCLabs.KitServer.2/{c8a12f17-1e03-401e-b53d-6c654dd576da}",
                "Simulation.Constant_BOOL",
                typeof(Boolean));
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteItemValueCatalogTests_Basic()
        {
            var tests = new List<Action>();
            foreach (DAServerCatalogEntry serverCatalogEntry in ServerCatalog)
            {
                Debug.Assert(serverCatalogEntry != null);

                IEnumerable<DAItemCatalogEntry> itemCatalogEntries = serverCatalogEntry.GetReadAndWriteableItems();
                foreach (DAItemCatalogEntry itemCatalogEntry in itemCatalogEntries)
                {
                    Debug.Assert(itemCatalogEntry != null);

                    DAServerCatalogEntry server = serverCatalogEntry;
                    DAItemCatalogEntry item = itemCatalogEntry;
                    tests.Add(() => WriteItemValueTest_Basic(server.Element, item.Element));
                }
            }
            TestUtilities.RunTests(tests);
        }

        #endregion



        #region Protected Methods

        protected void BrowseNodesTest_Basic([NotNull] ServerDescriptor serverDescriptor, [NotNull] DANodeDescriptor parentNodeDescriptor)
        {
            TestEnvironment.PrecheckServer(serverDescriptor);
            IEasyDAClient client = CreateClient();
            DANodeElementCollection nodeElements;
            try
            {
                nodeElements = client.BrowseNodes(
                    serverDescriptor,
                    parentNodeDescriptor,
                    DABrowseParameters.Empty);
            }
            catch (OpcException opcException)
            {
                if (!client.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            Assert.IsNotNull(nodeElements);
            AssertUtilities.WriteLine(nodeElements.Count);
            CollectionAssert.AllItemsAreNotNull(nodeElements);
        }

        protected void GetPropertyValueTest_Basic(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] DAItemDescriptor itemDescriptor,
            [NotNull] DAPropertyDescriptor propertyDescriptor,
            [CanBeNull] Type expectedType)
        {
            AssertUtilities.WriteLine(serverDescriptor);
            AssertUtilities.WriteLine(itemDescriptor);
            AssertUtilities.WriteLine(propertyDescriptor);
            AssertUtilities.WriteLine("");

            TestEnvironment.PrecheckServer(serverDescriptor);
            IEasyDAClient client = CreateClient();
            object actualValue;
            try
            {
                actualValue = client.GetPropertyValue(serverDescriptor, itemDescriptor, propertyDescriptor);
            }
            catch (OpcException opcException)
            {
                if (!client.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            AssertUtilities.WriteLine(actualValue);

            if (expectedType != null)
            {
                Assert.IsNotNull(actualValue);
                Assert.AreEqual(expectedType, actualValue.GetType());
            }
        }

        protected void GetPropertyValueTest_Timestamp(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] DAItemDescriptor itemDescriptor,
            [NotNull] DAPropertyDescriptor propertyDescriptor)
        {
            AssertUtilities.WriteLine(serverDescriptor);
            AssertUtilities.WriteLine(itemDescriptor);
            AssertUtilities.WriteLine(propertyDescriptor);
            AssertUtilities.WriteLine("");

            TestEnvironment.PrecheckServer(serverDescriptor);
            IEasyDAClient client = CreateClient();
            object actualValue;
            try
            {
                actualValue = client.GetPropertyValue(serverDescriptor, itemDescriptor, propertyDescriptor);
            }
            catch (OpcException opcException)
            {
                if (!client.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            AssertUtilities.WriteLine(actualValue);

            Assert.IsNotNull(actualValue);
            Assert.IsInstanceOfType(actualValue, typeof(DateTime));
            var dateTime = (DateTime) actualValue;
            AssertUtilities.WriteLine(dateTime.Kind);

            AssertUtilities.WriteLine(serverDescriptor.Technology);
            if (serverDescriptor.Technology == OpcTechnologies.Com)
                Assert.AreEqual(DateTimeKind.Utc, dateTime.Kind);
            // OPC-XML servers can return any kind here
        }

        protected void ReadItemTest_Basic(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] DAItemDescriptor itemDescriptor,
            [CanBeNull] Type expectedType)
        {
            AssertUtilities.WriteLine(serverDescriptor);
            AssertUtilities.WriteLine(itemDescriptor);
            AssertUtilities.WriteLine("");

            TestEnvironment.PrecheckServer(serverDescriptor);
            IEasyDAClient client = CreateClient();
            DAVtq actualVtq;
            try
            {
                actualVtq = client.ReadItem(serverDescriptor, itemDescriptor);
            }
            catch (OpcException opcException)
            {
                if (!client.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }

            Assert.IsNotNull(actualVtq);
            AssertUtilities.WriteLine(actualVtq);

            if (expectedType != null)
            {
                object actualValue = actualVtq.Value;
                Assert.IsNotNull(actualValue);
                Assert.AreEqual(expectedType, actualValue.GetType());
            }

            Assert.AreEqual(DateTimeKind.Utc, actualVtq.Timestamp.Kind);
            Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MinValue);
            Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MaxValue);
            Assert.IsTrue(actualVtq.Timestamp >= DateTime.FromFileTimeUtc(0));
            Assert.IsTrue(actualVtq.Timestamp <= DateTime.Now);
        }

        protected void ReadItemTest_Constant_Basic(
            [NotNull] ServerDescriptor serverDescriptor, 
            [NotNull] DAItemDescriptor itemDescriptor,
            [NotNull] DAReadParameters readParameters,
            [NotNull] DAQuality expectedQuality,
            [CanBeNull] object expectedValue)
        {
            AssertUtilities.WriteLine(serverDescriptor);
            AssertUtilities.WriteLine(itemDescriptor);
            AssertUtilities.WriteLine("");

            TestEnvironment.PrecheckServer(serverDescriptor);
            IEasyDAClient client = CreateClient();
            DAVtq actualVtq;
            try
            {
                actualVtq = client.ReadItem(serverDescriptor, itemDescriptor, readParameters);
            }
            catch (OpcException opcException)
            {
                if (!client.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }

            Assert.IsNotNull(actualVtq);
            AssertUtilities.WriteLine(actualVtq);

            Assert.AreEqual(expectedQuality, actualVtq.Quality);
            Assert.AreEqual(DAStatusChoice.GoodNonspecific, actualVtq.Quality.StatusBitField);
            Assert.AreEqual(DALimitChoice.LimitOk, actualVtq.Quality.LimitBitField);

            Assert.AreEqual(DateTimeKind.Utc, actualVtq.Timestamp.Kind);
            Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MinValue);
            Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MaxValue);
            Assert.IsTrue(actualVtq.Timestamp >= DateTime.FromFileTimeUtc(0));
            Assert.IsTrue(actualVtq.Timestamp <= DateTime.Now);

            Assert.AreEqual(expectedValue, actualVtq.Value);
        }

        protected void ReadItemTest_NonReadable(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] DAItemDescriptor itemDescriptor)
        {
            AssertUtilities.WriteLine(serverDescriptor);
            AssertUtilities.WriteLine(itemDescriptor);
            AssertUtilities.WriteLine("");

            IEasyDAClient client = CreateClient();
            TestEnvironment.PrecheckServer(serverDescriptor);

            try
            {
                client.ReadItem(serverDescriptor, itemDescriptor);
            }
            catch (OpcException opcException)
            {
                AssertUtilities.WriteLine(opcException);
                // IMPROVE: test details of the exception
                return;
            }
            Assert.Fail();
        }

        protected void SubscribeItemTest_Basic(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] DAItemDescriptor itemDescriptor,
            [CanBeNull] Type expectedType)
        {
            AssertUtilities.WriteLine(serverDescriptor);
            AssertUtilities.WriteLine(itemDescriptor);
            AssertUtilities.WriteLine(expectedType);
            AssertUtilities.WriteLine("");

            TestEnvironment.PrecheckServer(serverDescriptor);
            IEasyDAClient client = CreateClient();

            var senderList = new List<object>();
            var argsList = new List<EasyDAItemChangedEventArgs>();
            int indexSet = -1;
            var completeEvent = new AutoResetEvent(initialState: false);

            void ClientOnItemChanged(object sender, EasyDAItemChangedEventArgs args)
            {
                AssertUtilities.WriteLine(args);
                senderList.Add(sender);
                argsList.Add(args);

                if (args != null)
                {
                    if ((args.Exception != null) || ((args.Vtq != null) && !args.Vtq.Quality.IsBad))
                    {
                        completeEvent.Set();
                        indexSet = argsList.Count - 1;
                    }
                }
            }

            client.ItemChanged += ClientOnItemChanged;
            const int requestedUpdateRate = 100;
            const int testState = 777777777;
            int handle = client.SubscribeItem(serverDescriptor, itemDescriptor, requestedUpdateRate, testState);

            bool completed = completeEvent.WaitOne(5*1000);
            client.UnsubscribeItem(handle);
            client.ItemChanged -= ClientOnItemChanged;

            int notificationCount = argsList.Count;
            AssertUtilities.WriteLine($"notificationCount: {notificationCount}");
            Assert.IsTrue(completed, "Not completed.");

            if (notificationCount == 1)
            {
                EasyDAItemChangedEventArgs args = argsList[0];
                // ReSharper disable once MergeSequentialChecks
                if ((args != null) && (args.Exception is NotImplementedException))
                    return; // test success
            }

            for (int i = 0; i < notificationCount; i++)
            {
                object sender = senderList[i];
                Assert.IsNotNull(sender);
                Assert.AreSame(client, sender);

                EasyDAItemChangedEventArgs args = argsList[i];
                Assert.IsNotNull(args);

                Assert.AreEqual(testState, args.Arguments.State); // is correct for value types only
                Assert.IsNull(args.Arguments.Callback);

                Assert.AreEqual(new DAGroupParameters(requestedUpdateRate), args.Arguments.GroupParameters);
                Assert.AreEqual(itemDescriptor, args.Arguments.ItemDescriptor);
                Assert.AreEqual(serverDescriptor, args.Arguments.ServerDescriptor);

                if (!client.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsFalse(args.Succeeded);
                    Assert.IsInstanceOfType(args.Exception, typeof(NotSupportedException));
                    continue;
                }

                Assert.IsNull(args.Exception);
                Assert.AreEqual(null, args.ErrorId);

                DAVtq actualVtq = args.Vtq;
                Assert.IsNotNull(actualVtq);
                AssertUtilities.WriteLine(actualVtq);

                object actualValue = actualVtq.Value;

                if ((indexSet == -1) || (i < indexSet))
                {
                    Assert.IsTrue(actualVtq.Quality.IsBad);
                    Assert.IsNull(actualValue);
                }
                else
                {
                    Assert.IsTrue(actualVtq.Quality.IsGood);
                    if (expectedType != null)
                    {
                        Assert.IsNotNull(actualValue);
                        Assert.AreEqual(expectedType, actualValue.GetType());
                    }
                }

                Assert.AreEqual(DateTimeKind.Utc, actualVtq.Timestamp.Kind);
                Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MinValue);
                Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MaxValue);
                Assert.IsTrue(actualVtq.Timestamp >= DateTime.FromFileTimeUtc(0));
                Assert.IsTrue(actualVtq.Timestamp <= DateTime.Now);
            }
        }

        protected void SubscribeItemTest_Error(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] DAItemDescriptor itemDescriptor)
        {
            AssertUtilities.WriteLine(serverDescriptor);
            AssertUtilities.WriteLine(itemDescriptor);
            AssertUtilities.WriteLine("");

            TestEnvironment.PrecheckServer(serverDescriptor);
            IEasyDAClient client = CreateClient();

            var senderList = new List<object>();
            var argsList = new List<EasyDAItemChangedEventArgs>();
            var completeEvent = new AutoResetEvent(initialState: false);

            void ClientOnItemChanged(object sender, EasyDAItemChangedEventArgs args)
            {
                AssertUtilities.WriteLine(args);
                senderList.Add(sender);
                argsList.Add(args);

                completeEvent.Set();
            }

            client.ItemChanged += ClientOnItemChanged;
            const int requestedUpdateRate = 100;
            const int testState = 777777777;
            int handle = client.SubscribeItem(serverDescriptor, itemDescriptor, requestedUpdateRate, testState);

            bool completed = completeEvent.WaitOne(5 * 1000);
            client.UnsubscribeItem(handle);
            client.ItemChanged -= ClientOnItemChanged;

            AssertUtilities.WriteLine(argsList.Count);
            Assert.IsTrue(completed, "Not completed.");
            Assert.AreEqual(1, argsList.Count);

            {
                object sender = senderList[0];
                Assert.IsNotNull(sender);
                Assert.AreSame(client, sender);

                EasyDAItemChangedEventArgs args = argsList[0];
                Assert.IsNotNull(args);

                Assert.AreEqual(testState, args.Arguments.State); // is correct for value types only
                Assert.IsNull(args.Arguments.Callback);

                Assert.AreEqual(new DAGroupParameters(requestedUpdateRate), args.Arguments.GroupParameters);
                Assert.AreEqual(itemDescriptor, args.Arguments.ItemDescriptor);
                Assert.AreEqual(serverDescriptor, args.Arguments.ServerDescriptor);

                Exception exception = args.Exception;
                Assert.IsNotNull(exception);
                AssertUtilities.WriteLine(exception);
                AssertUtilities.WriteLine(args.ErrorId ?? "");
                Assert.IsNull(args.Vtq);

                // IMPROVE: test more about the exception
            }
        }

        protected void WriteItemValueTest_Basic(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] DAItemDescriptor itemDescriptor)
        {
            AssertUtilities.WriteLine(serverDescriptor);
            AssertUtilities.WriteLine(itemDescriptor);
            AssertUtilities.WriteLine("");

            TestEnvironment.PrecheckServer(serverDescriptor);
            IEasyDAClient client = CreateClient();
            object value = null;
            if (client.SupportsTechnology(OpcTechnologies.Com))
                value = client.ReadItemValue(serverDescriptor, itemDescriptor);
            AssertUtilities.WriteLine(value);
            try
            {
                client.WriteItemValue(serverDescriptor, itemDescriptor, value);
            }
            catch (OpcException opcException)
            {
                if (!client.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
        }

        #endregion



        #region Private Properties

        [NotNull]
        private DALiveTestEnvironment TestEnvironment => DALiveTestEnvironment.Current;

        #endregion
    }
}
