﻿using System;
using System.Runtime.InteropServices;
using System.Threading;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib;
using OpcLabs.BaseLib.Execution;
using OpcLabs.BaseLib.Net.Sockets;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.AddressSpace;
using OpcLabs.EasyOpc.DataAccess.OperationModel;
using OpcLabs.EasyOpc.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.DataAccess
{
    [TestClass]

    public partial class EasyDAClientTest
    {
        protected class WcfClientStaticEnvironment : StaticEnvironment
        {
            //public WcfClientStaticEnvironment()
            //{
            //}
        }



        #region Static Public Methods

        #endregion



        #region Static Private Methods

        static private void WcfClientPrecheck()
        {
            EndpointDescriptor endpointDescriptor = EasyDAClient.ServiceProviderEndpointDescriptor;
            if (!TcpClientUtilities.CanCreate(endpointDescriptor))
                Assert.Inconclusive(
                    $"Failed pre-check: cannot create TCP client to service provider endpoint \"{endpointDescriptor}\".");
        }

        #endregion



        [Timeout(60 * 1000)]
        [TestCategory("Live")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void PullMultipleItemChangesTest_SubscribeItem()
        {
            var client = new EasyDAClient {PullItemChangedQueueCapacity = 1000};

            int handle = client.SubscribeItem("", "OPCLabs.KitServer.2", "Demo.Ramp", 1000);

            int count = 0;
            int startTickCount = System.Environment.TickCount;
            do
            {
                EasyDAItemChangedEventArgs[] eventArgsArray =
                    client.PullMultipleItemChanges(maximumCount:int.MaxValue, millisecondsTimeout: 2 * 1000);
                if (eventArgsArray == null)
                    continue;
                foreach (EasyDAItemChangedEventArgs eventArgs in eventArgsArray)
                {
                    AssertUtilities.WriteLine(eventArgs);
                    count++;

                    Assert.IsNotNull(eventArgs);
                    Exception exception = eventArgs.Exception;
                    Assert.IsNull(exception);
                    // IMPROVE: assert more about the event
                }
            } while (System.Environment.TickCount < startTickCount + 10 * 1000);

            client.UnsubscribeItem(handle);

            AssertUtilities.WriteLine(count);
            Assert.IsTrue(count >= 1);
        }

        [ExpectedException(typeof(InvalidOperationException))]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void PullMultipleItemChangesTest_ZeroCapacity()
        {
            var client = new EasyDAClient();
            client.PullMultipleItemChanges(maximumCount: int.MaxValue, millisecondsTimeout: 2000);
        }



        #region Test Methods

        [Ignore] // IMPROVE: currently, can only run as a solo test
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ClientParametersTest_ExecutionContext_InprocServer()
        {
            InternalClientParametersTest_ExecutionContext_InprocServer();
        }

        // this does not work yet, because static params cannot be transferred to web service.
        // IMPROVE: make the ExecutionContext part of the ServerDescriptor, or at least instance/adaptable params
        //[TestCategory("Live")]
        //[TestMethod]
        //public void ClientParametersTest_ExecutionContext_InprocServer_WcfClient()
        //{
        //    WcfClientPrecheck();
        //    using (new WcfClientStaticEnvironment())
        //        InternalClientParametersTest_ExecutionContext_InprocServer();
        //}

        [Ignore]    // IMPROVE: put back after making clear how the analyzer should be configured for this
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseLeavesTest_OpcAnalyzer_ProgId()
        {
            var target = new EasyDAClient();

            DANodeElementCollection actualResultCollection = target.BrowseNodes(
                new ServerDescriptor("OpcTestLab.OpcAnalyzer.1"),
                "",
                new DABrowseParameters(DABrowseFilter.All));
            Assert.IsNotNull(actualResultCollection);
            // IMPROVE: assert more
        }

        [Ignore] // IMPROVE: fails
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseNodesTest_BrowseNodesChunk_5()
        {
            BrowseNodesTest_BrowseNodesChunk(5);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_UnknownResource()
        {
            InternalReadItemTest_UnknownResource();
        }

        [Ignore] // IMPROVE: put back 
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_UnknownResource_WcfClient()
        {
            WcfClientPrecheck();
            using (new WcfClientStaticEnvironment())
                InternalReadItemTest_UnknownResource();
        }

        //[Ignore] // IMPROVE: put back 
        //[TestCategory("Live")]
        //[TestCategory("OpcClassic")]
        //[TestCategory("OpcDataAccess")]
        //[TestMethod]
        //public void ReadItemTest_UnregisteredProgID_WcfClient()
        //{
        //    WcfClientPrecheck();
        //    using (new WcfClientStaticEnvironment())
        //        InternalReadItemTest_UnregisteredProgID();
        //}

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest()
        {
            ReadItemTest("");
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_Remote()
        {
            ReadItemTest("T-OPC");
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void ReadItemTest_WaitThusDisconnect()
        {
            ReadItemTest("");
            Thread.Sleep(60 * 1000);
            // IMPROVE: test that it is truly disconnected now
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void ReadItemTest_WaitThusDisconnect_TOPServer()
        {
            ReadItemTest("", "SWToolbox.TOPServer.V6", "Simulation.Analog.Analog_1");
            Thread.Sleep(60 * 1000);
            // IMPROVE: test that it is truly disconnected now
        }


        [Ignore] // IMPROVE: put back
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_WcfClient()
        {
            WcfClientPrecheck();
            using (new WcfClientStaticEnvironment())
                ReadItemTest("");
        }

        #endregion



        #region Private Methods

        private void BrowseNodesTest_BrowseNodesChunk(int browseNodesChunk)
        {
            using (new StaticEnvironment())
            {
                EasyDAClient.SharedParameters.ClientParameters.BrowseNodesChunk = browseNodesChunk;

                var target = new EasyDAClient();

                DANodeElementCollection actualResultCollection = target.BrowseNodes(
                    new ServerDescriptor(LegacyTestDefs.ProgIds.KitServer),
                    new DANodeDescriptor("Simulation"),
                    DABrowseParameters.Empty);

                Assert.IsNotNull(actualResultCollection);
                int count = actualResultCollection.Count;
                Assert.IsTrue(count >= 163);

                //TODO:
            }
        }

        private void InternalClientParametersTest_ExecutionContext_InprocServer()
        {
            TestSection.Arrange();
            using (new StaticEnvironment())
            {
                EasyDAClient.SharedParameters.ClientParameters.ExecutionContext = 1;
                var client = new EasyDAClient { Isolated = true };
                object value;
                TestSection.Act();
                try
                {
                    value = client.ReadItem("", LegacyTestDefs.ProgIds.KitServer,
                        LegacyTestDefs.DataAccess.ItemIds.Constant_I2);
                }
                catch (OpcException opcException)
                {
                    TestSection.Assert();
                    AssertUtilities.WriteLine(opcException);
                    Exception innerException = opcException.InnerException;
                    Assert.IsNotNull(innerException);
                    var comException = innerException as COMException;
                    Assert.IsNotNull(comException);
                    Assert.AreEqual(unchecked((int)0x80040154 /*Class not registered*/), comException.ErrorCode);
                    return;
                }
                AssertUtilities.WriteLine(value);
                TestSection.Assert();
            }
            Assert.Fail("OpcException not thrown.");
        }

        private void InternalReadItemTest_UnknownResource()
        {
            try
            {
                IEasyDAClient target = new EasyDAClient();
                target.ReadItem(LegacyTestDefs.MachineNames.UnknownResource, LegacyTestDefs.ProgIds.KitServer,
                             LegacyTestDefs.DataAccess.ItemIds.Constant_R4);
                Assert.Fail();
            }
            catch (OpcException ex)
            {
                Exception innerException = ex.InnerException;
                Assert.IsNotNull(innerException);
                var comException = innerException as COMException;
                Assert.IsNotNull(comException);
                int errorCode = comException.ErrorCode;
                AssertUtilities.WriteLine(errorCode);
                Assert.IsTrue(errorCode == -2147023174 || errorCode == -1073430527 || errorCode == -1073442813 ||
                              errorCode == (unchecked((int)0X80040154)));
                // The RPC server is unavailable OR Cannot connect client OR Machine not connected OR Class not registered
            }
        }

        private void ReadItemTest([NotNull] string machineName)
        {
            ReadItemTest(machineName, "OPCLabs.KitServer.2", "Demo.Ramp");
        }

        private void ReadItemTest([NotNull] string machineName, [NotNull] string objectId, [NotNull] string itemId)
        {
            var serverDescriptor = new ServerDescriptor(machineName, objectId);
            TestEnvironment.Precheck(serverDescriptor);

            var client = new EasyDAClient();

            var item = new DAItemDescriptor(itemId);

            DAVtq vtq = client.ReadItem(serverDescriptor, item);
            AssertUtilities.WriteLine(vtq);
        }

        #endregion
    }
}
