﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.ComponentModel.Linking;
using OpcLabs.EasyOpc.DataAccess.Generic;
using OpcLabs.EasyOpc.DataAccess.LiveMapping;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.LiveMapping._Testers
{
    class EventVtqTester
    {
        [DAType, DASubscription(RequestedUpdateRate = 100)]
        private class EventVtqTestType
        {
#pragma warning disable 67
            [DANode(ItemId = LegacyTestDefs.DataAccess.ItemIds.Incrementing_1s)]
            [DAItem(ItemType = typeof(Int32), Kind = DAItemMappingKind.Vtq)]
            public event TargetDataEventHandler<DAVtq<Int32>, EventVtqTestType> IncrementingChanged;
#pragma warning restore 67
        }

        public void Test()
        {
            const int timeout = 10 * 1000;
            const int period = 1000;

            int count = 0;

            var clientMapper = new DAClientMapper();
            var testTarget = new EventVtqTestType();
            // ReSharper disable PossibleNullReferenceException
            testTarget.IncrementingChanged += ((_, args) => Console.WriteLine(args.Data.ToString()));
            testTarget.IncrementingChanged += ((_, args) => count++);
            // ReSharper restore PossibleNullReferenceException

            clientMapper.Map(testTarget, new DAMappingContext { ServerDescriptor = LegacyTestDefs.ProgIds.KitServer });

            clientMapper.Subscribe(true);
            AssertUtilities.WaitingPhase(timeout);
            clientMapper.Subscribe(false);

            Assert.IsTrue(count >= timeout / period / 2);
        }
    }
}
