﻿using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Net.Sockets;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.Discovery;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Discovery
{
    public class UADiscoveryElementCollectionTest
    {
        #region Test Methods

        [TestCategory("BinarySerialization")]
        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void BinarySerializationTestLive()
        {
            Tests.BinarySerializationTest(CreateSerializationTargetLive());
        }

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestCategory("XmlSerialization")]
        [TestMethod]
        public void XmlSerializationTestLive_Localhost()
        {
            Tests.XmlSerializationTest(CreateSerializationTargetLive());
        }

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestCategory("XmlSerialization")]
        [TestMethod]
        public void XmlSerializationTestLive_Demo()
        {
            Tests.XmlSerializationTest(CreateSerializationTargetLive(LegacyTestDefs.MachineNames.OpcUaTest));
        }

        #endregion



        #region Private Methods

        [NotNull]
        private UADiscoveryElementCollection CreateSerializationTargetLive([NotNull] string discoveryHost = "localhost")
        {
            AssertUtilities.WriteLine(discoveryHost);
            if (!TcpClientUtilities.CanCreate(discoveryHost, 4840))
                Assert.Inconclusive("Failed pre-check: Socket exception.");

            return EasyUAClient.SharedInstance.DiscoverLocalServers(discoveryHost);
        }

        #endregion
    }
}
