﻿using System;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.AddressSpace;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpcLiveTest.Legacy.UA.Environment;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA
{
    class SeparationTester
    {
        private const int Duration = 10 * 1000;   // IMPROVE
        const int SamplingInterval = 100;

        private class SeparationState
        {
            public int Target1NotificationCount;
            public int Target2NotificationCount;
        }

        public void Test()
        {
            var target1 = new EasyUAClient();
            var target2 = new EasyUAClient();
            UAEndpointDescriptor endpointDescriptor = LegacyUATestEnvironment.OpcFoundation_SampleServer.EndpointDescriptors.Http;
            target1.DataChangeNotification += target1_DataChangeNotification;
            target2.DataChangeNotification += target2_DataChangeNotification;
            var separationState = new SeparationState();
            _target1Exception = null;
            _target2Exception = null;
            target1.SubscribeDataChange(endpointDescriptor, _separationNodeId1, SamplingInterval, separationState);
            target2.SubscribeDataChange(endpointDescriptor, _separationNodeId2, SamplingInterval, separationState);
            AssertUtilities.WaitingPhase(Duration);
            target1.UnsubscribeAllMonitoredItems();
            target2.UnsubscribeAllMonitoredItems();
            Assert.IsNull(_target1Exception);
            Assert.IsNull(_target2Exception);
            AssertUtilities.WriteLine(separationState.Target1NotificationCount);
            AssertUtilities.WriteLine(separationState.Target2NotificationCount);
            Assert.IsTrue(separationState.Target1NotificationCount >= 1);
            Assert.IsTrue(separationState.Target2NotificationCount >= 1);
        }

        // ReSharper disable InconsistentNaming
        private void target1_DataChangeNotification(object sender, EasyUADataChangeNotificationEventArgs eventArgs)
        // ReSharper restore InconsistentNaming
        {
            try
            {
                Assert.IsNotNull(eventArgs);
                object state = eventArgs.Arguments.State;
                Assert.IsNotNull(state);
                ((SeparationState)state).Target1NotificationCount++;
                Assert.AreEqual(eventArgs.Arguments.NodeDescriptor.NodeId, _separationNodeId1);
            }
            catch (Exception exception)
            {
                if (_target1Exception == null)
                    _target1Exception = exception;
            }
        }

        // ReSharper disable InconsistentNaming
        private void target2_DataChangeNotification(object sender, EasyUADataChangeNotificationEventArgs eventArgs)
        // ReSharper restore InconsistentNaming
        {
            try
            {
                Assert.IsNotNull(eventArgs);
                object state = eventArgs.Arguments.State;
                Assert.IsNotNull(state);
                ((SeparationState)state).Target2NotificationCount++;
                Assert.AreEqual(eventArgs.Arguments.NodeDescriptor.NodeId, _separationNodeId2);
            }
            catch (Exception exception)
            {
                if (_target2Exception == null)
                    _target2Exception = exception;
            }
        }

        [NotNull]
        private readonly UANodeId _separationNodeId1 = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Dynamic_Scalar_SByteValue;
        [NotNull]
        private readonly UANodeId _separationNodeId2 = OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Dynamic_Scalar_FloatValue;

        [CanBeNull]
        private Exception _target1Exception;
        [CanBeNull]
        private Exception _target2Exception;

    }
}
