﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib;
using OpcLabs.BaseLib.Extensions;
using OpcLabs.BaseLib.Instrumentation.Internal;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.Implementation;
using OpcLabs.EasyOpc.DataAccess.Internal;
using OpcLabs.EasyOpc.DataAccess.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcTest.DataAccess.Implementation
{
    [TestClass]
    public class StateLockingEasyDAClientTest
    {
        [TestMethod]
        public void Test()
        {
            TestSection.Arrange();

            var apiMonitoringEntryList = new List<ApiMonitoringEntry>();

            // ReSharper disable once ParameterOnlyUsedForPreconditionCheck.Local
            void MonitorApiHandler(object sender, DataEventArgs<ApiMonitoringEntry> eventArgs, int level)
            {
                Assert.IsNotNull(sender);
                Assert.IsNotNull(eventArgs);
                ApiMonitoringEntry apiMonitoringEntry = eventArgs.Data;
                Assert.IsNotNull(apiMonitoringEntry);
                AssertUtilities.WriteLine($"({apiMonitoringEntryList.Count}) {new String(' ', level*4)}{apiMonitoringEntry}");
                apiMonitoringEntryList.Add(apiMonitoringEntry);
            }
            
            var simulatedClient = new SimulatedEasyDAClient();
            var simulatedClientApiMonitoringClient = new ApiMonitoringEasyDAClient(simulatedClient);
            simulatedClientApiMonitoringClient.MonitorApi += (sender, args) => MonitorApiHandler(sender, args, level: 1);

            var stateLockingClient = new StateLockingEasyDAClient(simulatedClientApiMonitoringClient);
            var target = new ApiMonitoringEasyDAClient(stateLockingClient);
            target.MonitorApi += (sender, args) => MonitorApiHandler(sender, args, level: 0);
            
            AssertUtilities.WriteLine();
            IStateLocking innerStateLocking = target.GetService<IStateLocking>();
            Assert.IsNotNull(innerStateLocking);
            var stateLocking = new ApiMonitoringStateLocking(innerStateLocking);
            stateLocking.MonitorApi += (sender, args) => MonitorApiHandler(sender, args, level: 0);

            TestSection.Act();

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client unlocked; subscribe to item 1.");
            apiMonitoringEntryList.Clear();
            EasyDAItemSubscriptionArguments itemSubscriptionArguments1 = new EasyDAItemSubscriptionArguments("s1", "i1", 1, "ItemSubscription1");
            int handle1 = target.SubscribeItem(itemSubscriptionArguments1);
            AssertUtilities.WriteLine($"{nameof(handle1)}: {handle1}");
            // ReSharper disable PossibleNullReferenceException
            Assert.AreSame(stateLockingClient, apiMonitoringEntryList[0].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.SubscribeMultipleItems", apiMonitoringEntryList[0].TypedMemberName);
            Assert.AreEqual(1, ((EasyDAItemSubscriptionArguments[])apiMonitoringEntryList[0].Parameters[0]).Length);
            Assert.AreEqual(itemSubscriptionArguments1, ((EasyDAItemSubscriptionArguments[])apiMonitoringEntryList[0].Parameters[0])[0]);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[1].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.SubscribeMultipleItems", apiMonitoringEntryList[1].TypedMemberName);
            Assert.AreEqual(1, ((EasyDAItemSubscriptionArguments[])apiMonitoringEntryList[1].Parameters[0]).Length);
            Assert.AreEqual(itemSubscriptionArguments1.ServerDescriptor, ((EasyDAItemSubscriptionArguments[])apiMonitoringEntryList[1].Parameters[0])[0].ServerDescriptor);
            Assert.AreEqual(itemSubscriptionArguments1.ItemDescriptor, ((EasyDAItemSubscriptionArguments[])apiMonitoringEntryList[1].Parameters[0])[0].ItemDescriptor);
            Assert.AreEqual(itemSubscriptionArguments1.GroupParameters, ((EasyDAItemSubscriptionArguments[])apiMonitoringEntryList[1].Parameters[0])[0].GroupParameters);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[2].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ItemChanged", apiMonitoringEntryList[2].TypedMemberName);
            Assert.AreSame(stateLockingClient, apiMonitoringEntryList[3].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ItemChanged", apiMonitoringEntryList[3].TypedMemberName);
            Assert.AreSame(target, apiMonitoringEntryList[3].Parameters[0]);
            Assert.AreEqual(itemSubscriptionArguments1, ((EasyDAItemChangedEventArgs)apiMonitoringEntryList[3].Parameters[1]).Arguments);
            Assert.AreSame(stateLockingClient, apiMonitoringEntryList[4].TargetOrSourceObject);
            Assert.IsTrue(apiMonitoringEntryList[4].IsReturn);
            Assert.AreEqual("IEasyDAClient.ItemChanged", apiMonitoringEntryList[4].TypedMemberName);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[5].TargetOrSourceObject);
            Assert.IsTrue(apiMonitoringEntryList[5].IsReturn);
            Assert.AreEqual("IEasyDAClient.ItemChanged", apiMonitoringEntryList[5].TypedMemberName);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[6].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.SubscribeMultipleItems", apiMonitoringEntryList[6].TypedMemberName);
            Assert.IsTrue(apiMonitoringEntryList[6].IsReturn);
            int innerHandle1 = ((int[])apiMonitoringEntryList[6].ReturnValue)[0];
            AssertUtilities.WriteLine($"{nameof(innerHandle1)}: {innerHandle1}");
            Assert.AreSame(stateLockingClient, apiMonitoringEntryList[7].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.SubscribeMultipleItems", apiMonitoringEntryList[7].TypedMemberName);
            Assert.IsTrue(apiMonitoringEntryList[7].IsReturn);
            Assert.AreEqual(8, apiMonitoringEntryList.Count);
            // ReSharper restore PossibleNullReferenceException

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client unlocked; check that the returned handle is known by the client.");
            apiMonitoringEntryList.Clear();
            var handle1IsKnown = target.IsKnownItemSubscriptionHandle(handle1);
            Assert.IsTrue(handle1IsKnown);

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client unlocked; check that a fabricated handle is *not* known by the client.");
            apiMonitoringEntryList.Clear();
            int handle0 = handle1 - 1;
            AssertUtilities.WriteLine($"{nameof(handle0)}: {handle0}");
            var handle0IsKnown = target.IsKnownItemSubscriptionHandle(handle0);
            Assert.IsFalse(handle0IsKnown);

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client unlocked; get item subscription arguments for a known handle.");
            apiMonitoringEntryList.Clear();
            EasyDAItemSubscriptionArguments actualItemSubscriptionArguments1 = target.GetItemSubscriptionArguments(handle1);
            AssertUtilities.WriteLine($"{nameof(actualItemSubscriptionArguments1)}: {actualItemSubscriptionArguments1}");
            Assert.AreEqual(itemSubscriptionArguments1, actualItemSubscriptionArguments1);

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client unlocked; attempt to get item subscription arguments for an unknown handle.");
            apiMonitoringEntryList.Clear();
            Exception exception = null;
            try
            {
                // ReSharper disable once ReturnValueOfPureMethodIsNotUsed
                target.GetItemSubscriptionArguments(handle0);
            }
            catch (Exception e)
            {
                exception = e;
            }
            Assert.IsInstanceOfType(exception, typeof(ArgumentException));
            Assert.AreEqual(2, apiMonitoringEntryList.Count);

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client unlocked; change item subscription for a known handle.");
            apiMonitoringEntryList.Clear();
            var groupParameters1001 = new DAGroupParameters(1001);
            target.ChangeItemSubscription(handle1, groupParameters1001);
            // ReSharper disable PossibleNullReferenceException
            Assert.AreSame(stateLockingClient, apiMonitoringEntryList[0].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ChangeMultipleItemSubscriptions", apiMonitoringEntryList[0].TypedMemberName);
            Assert.AreEqual(1, ((DAHandleGroupArguments[])apiMonitoringEntryList[0].Parameters[0]).Length);
            var actualHandleGroupArguments1A = ((DAHandleGroupArguments[]) apiMonitoringEntryList[0].Parameters[0])[0];
            Assert.IsNotNull(actualHandleGroupArguments1A);
            Assert.AreEqual(handle1, actualHandleGroupArguments1A.Handle);
            Assert.AreEqual(groupParameters1001, actualHandleGroupArguments1A.GroupParameters);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[1].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ChangeMultipleItemSubscriptions", apiMonitoringEntryList[1].TypedMemberName);
            Assert.AreEqual(1, ((DAHandleGroupArguments[])apiMonitoringEntryList[1].Parameters[0]).Length);
            var actualHandleGroupArguments1B = ((DAHandleGroupArguments[])apiMonitoringEntryList[1].Parameters[0])[0];
            Assert.IsNotNull(actualHandleGroupArguments1B);
            Assert.AreEqual(innerHandle1, actualHandleGroupArguments1B.Handle);
            Assert.AreEqual(groupParameters1001, actualHandleGroupArguments1B.GroupParameters);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[2].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ChangeMultipleItemSubscriptions", apiMonitoringEntryList[2].TypedMemberName);
            Assert.IsTrue(apiMonitoringEntryList[2].IsReturn);
            Assert.AreSame(stateLockingClient, apiMonitoringEntryList[5].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ChangeMultipleItemSubscriptions", apiMonitoringEntryList[5].TypedMemberName);
            Assert.IsTrue(apiMonitoringEntryList[5].IsReturn);
            Assert.AreEqual(6, apiMonitoringEntryList.Count);
            // ReSharper restore PossibleNullReferenceException

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client unlocked; attempt to change item subscription for an unknown handle.");
            apiMonitoringEntryList.Clear();
            exception = null;
            try
            {
                target.ChangeItemSubscription(handle0, new DAGroupParameters());
            }
            catch (Exception e)
            {
                exception = e;
            }
            Assert.IsInstanceOfType(exception, typeof(ArgumentException));
            Assert.AreEqual(2, apiMonitoringEntryList.Count);

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client unlocked; attempt to to unsubscribe using an unknown subscription handle.");
            apiMonitoringEntryList.Clear();
            exception = null;
            try
            {
                target.UnsubscribeItem(handle0);
            }
            catch (Exception e)
            {
                exception = e;
            }
            Assert.IsInstanceOfType(exception, typeof(ArgumentException));
            Assert.AreEqual(2, apiMonitoringEntryList.Count);

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client unlocked; unsubscribe from an item.");
            apiMonitoringEntryList.Clear();
            target.UnsubscribeItem(handle1);
            // ReSharper disable PossibleNullReferenceException
            Assert.AreSame(stateLockingClient, apiMonitoringEntryList[0].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.UnsubscribeMultipleItems", apiMonitoringEntryList[0].TypedMemberName);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[1].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.UnsubscribeMultipleItems", apiMonitoringEntryList[1].TypedMemberName);
            var actualInnerHandles = (IEnumerable<int>)apiMonitoringEntryList[1].Parameters[0];
            Assert.IsNotNull(actualInnerHandles);
            Assert.IsTrue(actualInnerHandles.SequenceEqual(new[] {innerHandle1}));
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[2].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.UnsubscribeMultipleItems", apiMonitoringEntryList[2].TypedMemberName);
            Assert.IsTrue(apiMonitoringEntryList[2].IsReturn);
            Assert.AreSame(stateLockingClient, apiMonitoringEntryList[3].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.UnsubscribeMultipleItems", apiMonitoringEntryList[3].TypedMemberName);
            Assert.IsTrue(apiMonitoringEntryList[3].IsReturn);
            Assert.AreEqual(4, apiMonitoringEntryList.Count);
            // ReSharper restore PossibleNullReferenceException

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client unlocked; subscribe to item 2.");
            apiMonitoringEntryList.Clear();
            EasyDAItemSubscriptionArguments itemSubscriptionArguments2 = new EasyDAItemSubscriptionArguments("s2", "i2", 2, "ItemSubscription2");
            int handle2 = target.SubscribeItem(itemSubscriptionArguments2);
            AssertUtilities.WriteLine($"{nameof(handle2)}: {handle2}");

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client unlocked; unsubscribe from all items.");
            apiMonitoringEntryList.Clear();
            target.UnsubscribeAllItems();
            // ReSharper disable PossibleNullReferenceException
            Assert.AreSame(stateLockingClient, apiMonitoringEntryList[0].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.UnsubscribeAllItems", apiMonitoringEntryList[0].TypedMemberName);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[1].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.UnsubscribeAllItems", apiMonitoringEntryList[1].TypedMemberName);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[2].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.UnsubscribeAllItems", apiMonitoringEntryList[2].TypedMemberName);
            Assert.IsTrue(apiMonitoringEntryList[2].IsReturn);
            Assert.AreSame(stateLockingClient, apiMonitoringEntryList[3].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.UnsubscribeAllItems", apiMonitoringEntryList[3].TypedMemberName);
            Assert.IsTrue(apiMonitoringEntryList[3].IsReturn);
            Assert.AreEqual(4, apiMonitoringEntryList.Count);
            // ReSharper restore PossibleNullReferenceException

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client unlocked; subscribe to item 3.");
            apiMonitoringEntryList.Clear();
            EasyDAItemSubscriptionArguments itemSubscriptionArguments3 = new EasyDAItemSubscriptionArguments("s3", "i3", 3, "ItemSubscription3");
            int handle3 = target.SubscribeItem(itemSubscriptionArguments3);
            AssertUtilities.WriteLine($"{nameof(handle3)}: {handle3}");
            // ReSharper disable PossibleNullReferenceException
            int innerHandle3 = ((int[])apiMonitoringEntryList[6].ReturnValue)[0];
            AssertUtilities.WriteLine($"{nameof(innerHandle3)}: {innerHandle3}");
            // ReSharper restore PossibleNullReferenceException

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client unlocked; subscribe to item 4.");
            apiMonitoringEntryList.Clear();
            EasyDAItemSubscriptionArguments itemSubscriptionArguments4 = new EasyDAItemSubscriptionArguments("s4", "i4", 4, "ItemSubscription4");
            int handle4 = target.SubscribeItem(itemSubscriptionArguments4);
            AssertUtilities.WriteLine($"{nameof(handle4)}: {handle4}");
            // ReSharper disable PossibleNullReferenceException
            int innerHandle4 = ((int[])apiMonitoringEntryList[6].ReturnValue)[0];
            AssertUtilities.WriteLine($"{nameof(innerHandle4)}: {innerHandle4}");
            // ReSharper restore PossibleNullReferenceException

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client unlocked; subscribe to item 5.");
            apiMonitoringEntryList.Clear();
            EasyDAItemSubscriptionArguments itemSubscriptionArguments5 = new EasyDAItemSubscriptionArguments("s5", "i5", 5, "ItemSubscription5");
            int handle5 = target.SubscribeItem(itemSubscriptionArguments5);
            AssertUtilities.WriteLine($"{nameof(handle5)}: {handle5}");
            // ReSharper disable PossibleNullReferenceException
            int innerHandle5 = ((int[])apiMonitoringEntryList[6].ReturnValue)[0];
            AssertUtilities.WriteLine($"{nameof(innerHandle5)}: {innerHandle5}");
            // ReSharper restore PossibleNullReferenceException

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client unlocked; subscribe to item 6.");
            apiMonitoringEntryList.Clear();
            EasyDAItemSubscriptionArguments itemSubscriptionArguments6 = new EasyDAItemSubscriptionArguments("s6", "i6", 6, "ItemSubscription6");
            int handle6 = target.SubscribeItem(itemSubscriptionArguments6);
            AssertUtilities.WriteLine($"{nameof(handle6)}: {handle6}");
            // ReSharper disable PossibleNullReferenceException
            int innerHandle6 = ((int[])apiMonitoringEntryList[6].ReturnValue)[0];
            AssertUtilities.WriteLine($"{nameof(innerHandle6)}: {innerHandle6}");
            // ReSharper restore PossibleNullReferenceException

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Obtain state locked flag, check that it is unlocked.");
            apiMonitoringEntryList.Clear();
            bool isStateLocked0 = stateLocking.IsStateLocked;
            Assert.IsFalse(isStateLocked0);

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Lock the state.");
            apiMonitoringEntryList.Clear();
            stateLocking.LockState();

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Obtain state locked flag, check that it is locked.");
            apiMonitoringEntryList.Clear();
            bool isStateLocked1 = stateLocking.IsStateLocked;
            Assert.IsTrue(isStateLocked1);

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client locked; get subscription arguments for item 3.");
            apiMonitoringEntryList.Clear();
            EasyDAItemSubscriptionArguments actualItemSubscriptionArguments3 = target.GetItemSubscriptionArguments(handle3);
            AssertUtilities.WriteLine($"{nameof(actualItemSubscriptionArguments3)}: {actualItemSubscriptionArguments3}");
            Assert.AreEqual(itemSubscriptionArguments3, actualItemSubscriptionArguments3);

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client locked; unsubscribe from item 3.");
            apiMonitoringEntryList.Clear();
            target.UnsubscribeItem(handle3);
            Assert.AreEqual(2, apiMonitoringEntryList.Count);

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client locked; change item subscription 4.");
            apiMonitoringEntryList.Clear();
            DAGroupParameters groupParameters1004 = new DAGroupParameters(1004);
            target.ChangeItemSubscription(handle4, groupParameters1004);
            // ReSharper disable PossibleNullReferenceException
            Assert.AreSame(stateLockingClient, apiMonitoringEntryList[0].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ChangeMultipleItemSubscriptions", apiMonitoringEntryList[0].TypedMemberName);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[1].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.GetItemSubscriptionArguments", apiMonitoringEntryList[1].TypedMemberName);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[2].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.GetItemSubscriptionArguments", apiMonitoringEntryList[2].TypedMemberName);
            Assert.IsTrue(apiMonitoringEntryList[2].IsReturn);
            Assert.AreSame(stateLockingClient, apiMonitoringEntryList[5].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ChangeMultipleItemSubscriptions", apiMonitoringEntryList[5].TypedMemberName);
            Assert.IsTrue(apiMonitoringEntryList[5].IsReturn);
            Assert.AreEqual(6, apiMonitoringEntryList.Count);
            // ReSharper restore PossibleNullReferenceException

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client locked; get subscription arguments for item 4.");
            apiMonitoringEntryList.Clear();
            EasyDAItemSubscriptionArguments actualItemSubscriptionArguments4 = target.GetItemSubscriptionArguments(handle4);
            AssertUtilities.WriteLine($"{nameof(actualItemSubscriptionArguments4)}: {actualItemSubscriptionArguments4}");
            var expectedItemSubscriptionArguments4 = new EasyDAItemSubscriptionArguments(itemSubscriptionArguments4)
                {GroupParameters = groupParameters1004};
            Assert.AreEqual(expectedItemSubscriptionArguments4, actualItemSubscriptionArguments4);

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client locked; change item subscription 5.");
            apiMonitoringEntryList.Clear();
            DAGroupParameters groupParameters1005 = new DAGroupParameters(1005);
            target.ChangeItemSubscription(handle5, groupParameters1005);
            // ReSharper disable PossibleNullReferenceException
            Assert.AreSame(stateLockingClient, apiMonitoringEntryList[0].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ChangeMultipleItemSubscriptions", apiMonitoringEntryList[0].TypedMemberName);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[1].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.GetItemSubscriptionArguments", apiMonitoringEntryList[1].TypedMemberName);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[2].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.GetItemSubscriptionArguments", apiMonitoringEntryList[2].TypedMemberName);
            Assert.IsTrue(apiMonitoringEntryList[2].IsReturn);
            Assert.AreSame(stateLockingClient, apiMonitoringEntryList[5].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ChangeMultipleItemSubscriptions", apiMonitoringEntryList[5].TypedMemberName);
            Assert.IsTrue(apiMonitoringEntryList[5].IsReturn);
            Assert.AreEqual(6, apiMonitoringEntryList.Count);
            // ReSharper restore PossibleNullReferenceException

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client locked; revert the changes to item subscription 5.");
            apiMonitoringEntryList.Clear();
            target.ChangeItemSubscription(handle5, 5);
            // ReSharper disable PossibleNullReferenceException
            Assert.AreSame(stateLockingClient, apiMonitoringEntryList[0].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ChangeMultipleItemSubscriptions", apiMonitoringEntryList[0].TypedMemberName);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[1].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.GetItemSubscriptionArguments", apiMonitoringEntryList[1].TypedMemberName);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[2].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.GetItemSubscriptionArguments", apiMonitoringEntryList[2].TypedMemberName);
            Assert.IsTrue(apiMonitoringEntryList[2].IsReturn);
            Assert.AreSame(stateLockingClient, apiMonitoringEntryList[5].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ChangeMultipleItemSubscriptions", apiMonitoringEntryList[5].TypedMemberName);
            Assert.IsTrue(apiMonitoringEntryList[5].IsReturn);
            Assert.AreEqual(6, apiMonitoringEntryList.Count);
            // ReSharper restore PossibleNullReferenceException

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client locked; unsubscribe from item 6.");
            apiMonitoringEntryList.Clear();
            target.UnsubscribeItem(handle6);
            Assert.AreEqual(2, apiMonitoringEntryList.Count);

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client locked; subscribe to item 7 which has the same parameters as former item 6.");
            apiMonitoringEntryList.Clear();
            EasyDAItemSubscriptionArguments itemSubscriptionArguments7 = itemSubscriptionArguments6;
            int handle7 = target.SubscribeItem(itemSubscriptionArguments7);
            AssertUtilities.WriteLine($"{nameof(handle7)}: {handle7}");
            Assert.AreEqual(2, apiMonitoringEntryList.Count);

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client locked; subscribe to item 8.");
            apiMonitoringEntryList.Clear();
            EasyDAItemSubscriptionArguments itemSubscriptionArguments8 = new EasyDAItemSubscriptionArguments("s8", "i8", 8, "ItemSubscription8");
            int handle8 = target.SubscribeItem(itemSubscriptionArguments8);
            AssertUtilities.WriteLine($"{nameof(handle8)}: {handle8}");
            Assert.AreEqual(2, apiMonitoringEntryList.Count);

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client locked; make null change to item subscription 8.");
            apiMonitoringEntryList.Clear();
            target.ChangeItemSubscription(handle8, itemSubscriptionArguments8.GroupParameters);

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client locked; subscribe to item 9.");
            apiMonitoringEntryList.Clear();
            EasyDAItemSubscriptionArguments itemSubscriptionArguments9 = new EasyDAItemSubscriptionArguments("s9", "i9", 9, "ItemSubscription9");
            int handle9 = target.SubscribeItem(itemSubscriptionArguments9);
            AssertUtilities.WriteLine($"{nameof(handle9)}: {handle9}");
            Assert.AreEqual(2, apiMonitoringEntryList.Count);

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client locked; unsubscribe from item 8.");
            apiMonitoringEntryList.Clear();
            target.UnsubscribeItem(handle8);
            Assert.AreEqual(2, apiMonitoringEntryList.Count);

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Client locked; change item subscription 9.");
            apiMonitoringEntryList.Clear();
            var groupParameters1009 = new DAGroupParameters(1009);
            target.ChangeItemSubscription(handle9, groupParameters1009);
            Assert.AreEqual(2, apiMonitoringEntryList.Count);

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Unlock the state.");
            apiMonitoringEntryList.Clear();
            stateLocking.UnlockState();
            // ReSharper disable PossibleNullReferenceException
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[5].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.UnsubscribeMultipleItems", apiMonitoringEntryList[5].TypedMemberName);
            var actualHandlesToUnsubscribe = (IEnumerable<int>)apiMonitoringEntryList[5].Parameters[0];
            Assert.IsNotNull(actualHandlesToUnsubscribe);
            Assert.IsTrue(actualHandlesToUnsubscribe.SequenceEqual(new []{innerHandle3}));
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[7].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ChangeMultipleItemSubscriptions", apiMonitoringEntryList[7].TypedMemberName);
            var actualHandleGroupArgumentsArray = (DAHandleGroupArguments[]) apiMonitoringEntryList[7].Parameters[0];
            Assert.IsNotNull(actualHandleGroupArgumentsArray);
            Assert.IsTrue(actualHandleGroupArgumentsArray.SequenceEqual(new [] 
                {new DAHandleGroupArguments(innerHandle4, groupParameters1004)}));
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[9].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.SubscribeMultipleItems", apiMonitoringEntryList[9].TypedMemberName);
            var actualItemSubscriptionArgumentsArray = (EasyDAItemSubscriptionArguments[])apiMonitoringEntryList[9].Parameters[0];
            Assert.IsNotNull(actualItemSubscriptionArgumentsArray);
            Assert.AreEqual(1, actualItemSubscriptionArgumentsArray.Length);
            EasyDAItemSubscriptionArguments actualItemSubscriptionArguments = actualItemSubscriptionArgumentsArray[0];
            Assert.IsNotNull(actualItemSubscriptionArguments);
            // ReSharper disable once AssignNullToNotNullAttribute
            var actualItemSubscriptionArgumentsNoState = new EasyDAItemSubscriptionArguments(actualItemSubscriptionArguments) {State = null};
            Assert.AreEqual(
                new EasyDAItemSubscriptionArguments(itemSubscriptionArguments9) { GroupParameters = groupParameters1009, State = null},
                actualItemSubscriptionArgumentsNoState);
            Assert.AreEqual(16, apiMonitoringEntryList.Count);
            // ReSharper restore PossibleNullReferenceException

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Obtain state locked flag, check that it is unlocked.");
            apiMonitoringEntryList.Clear();
            bool isStateLocked2 = stateLocking.IsStateLocked;
            Assert.IsFalse(isStateLocked2);
        }
    }
}
