﻿using System.Collections;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.DataTypeModel;
using OpcLabs.EasyOpc.UA.DataTypeModel.Implementation;

namespace OpcLabs.EasyOpcTest.UA.DataTypeModel.Internal
{
    partial class TypeDescriptionExamplesTest
    {
        #region Static Public Properties

        [NotNull]
        static public OpaqueDataType Int128DataType =>
            new OpaqueDataType("Int128", "TypeDescriptionExamples:Int128", 128) { Description = "A 128-bit signed integer." };

        #endregion



        #region Test Methods

        [TestMethod]
        public void DecodeErrorTest_BinaryDecoderPrematureEndOfEncodedStream() =>
            DecodeErrorTest(null,
                TypeDictionaryByteArray,
                "Int128",
                new byte[0],
                new[] { UADataTypeModelErrorCodes.OpcBinarySystemDictionaryDecoderFailure, UADataTypeModelErrorCodes.OpcBinaryDictionaryDecoderPrematureEndOfEncodedStream});

        [TestMethod]
        public void DecodeTest_Int128_0() =>
            DecodeTest(null, TypeDictionaryByteArray, "Int128",
                new byte[16],
                new OpaqueData(Int128DataType, new BitArray(128)));

        [TestMethod]
        public void DecodeTest_Int128_1()
        {
            var bitArray = new BitArray(128) { [0] = true };
            var expectedOpaqueData = new OpaqueData(Int128DataType, bitArray);
            var extensionObjectBody = new byte[16];
            extensionObjectBody[0] = 1;
            DecodeTest(null, TypeDictionaryByteArray, "Int128",
                extensionObjectBody,
                expectedOpaqueData);
        }

        [TestMethod]
        public void EncodeErrorTest_BinaryEncoderEnumerationDataUnexpectedTypeDescriptor() =>
            EncodeErrorTest(null, 
                TypeDictionaryByteArray,
                "Int128",
                new EnumerationData(0),
                new[] { UADataTypeModelErrorCodes.OpcBinarySystemDictionaryEncoderFailure, UADataTypeModelErrorCodes.OpcBinaryDictionaryEncoderEnumerationDataUnexpectedTypeDescriptor});

        [TestMethod]
        public void EncodeErrorTest_BinaryEncoderSequenceDataOnlyAllowedInFields() =>
            EncodeErrorTest(null, 
                TypeDictionaryByteArray,
                "Int128",
                new SequenceData(),
                new[] { UADataTypeModelErrorCodes.OpcBinarySystemDictionaryEncoderFailure, UADataTypeModelErrorCodes.OpcBinaryDictionaryEncoderSequenceDataOnlyAllowedInFields});

        [TestMethod]
        public void EncodeErrorTest_BinaryEncoderOpaqueDataActualLengthInBitsMismatch() =>
            EncodeErrorTest(null, 
                TypeDictionaryByteArray,
                "Int128",
                new OpaqueData(new BitArray(129)),
                new[] { UADataTypeModelErrorCodes.OpcBinarySystemDictionaryEncoderFailure, UADataTypeModelErrorCodes.OpcBinaryDictionaryEncoderOpaqueDataActualLengthInBitsMismatch});

        [TestMethod]
        public void EncodeErrorTest_BinaryEncoderStructuredDataUnexpectedTypeDescriptor() =>
            EncodeErrorTest(null, 
                TypeDictionaryByteArray,
                "Int128",
                new StructuredData(),
                new[] { UADataTypeModelErrorCodes.OpcBinarySystemDictionaryEncoderFailure, UADataTypeModelErrorCodes.OpcBinaryDictionaryEncoderStructuredDataUnexpectedTypeDescriptor});

        [TestMethod]
        public void EncodeErrorTest_BinaryEncoderPrimitiveDataNamespaceNotRecognized() =>
            EncodeErrorTest(null, 
                TypeDictionaryByteArray,
                "Int128",
                new PrimitiveData(0),
                new[] { UADataTypeModelErrorCodes.OpcBinarySystemDictionaryEncoderFailure, UADataTypeModelErrorCodes.OpcBinaryDictionaryEncoderPrimitiveDataNamespaceNotRecognized});

        [TestMethod]
        public void EncodeTest_Int128_0()
        {
            var opaqueData = new OpaqueData(new BitArray(128));
            EncodeTest(null, TypeDictionaryByteArray, "Int128", opaqueData, new byte[16]);
        }

        [TestMethod]
        public void EncodeTest_Int128_1()
        {
            var bitArray = new BitArray(128) { [0] = true };
            var opaqueData = new OpaqueData(bitArray);
            var expectedExtensionObjectBody = new byte[16];
            expectedExtensionObjectBody[0] = 1;
            EncodeTest(null, TypeDictionaryByteArray, "Int128", opaqueData, expectedExtensionObjectBody);
        }

        [TestMethod]
        public void EncodeTest_Int128_Minus1()
        {
            var opaqueData = new OpaqueData(new BitArray(128, true));
            EncodeTest(null, TypeDictionaryByteArray, "Int128", opaqueData,
                new byte[] { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF });
        }

        #endregion
    }
}
