﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc.UA.PubSub;
using OpcLabs.EasyOpc.UA.PubSub.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.OpcInteroperabilityTest.UA.PubSub.Publishers
{
    [TestClass]
    public class SoftingIndustrialAutomation_CppSdkForWindows_Test
    {
        [TestMethod]
        public void DynamicLayoutTest_Physical()
        {
            UAPubSubConnectionDescriptor connectionDescriptor = "opc.udp://239.0.0.1";
            UASubscribeDataSetFilter subscribeDataSetFilter =
                UASubscribeDataSetFilter.Create(UAPublisherIdType.UInt16, 196);

            var subscriber = new EasyUASubscriber();
            subscriber.DataSetMessage += (sender, args) =>
            {
                AssertUtilities.WriteLine(args);
            };

            var subscribeDataSetArguments = new EasyUASubscribeDataSetArguments(connectionDescriptor, subscribeDataSetFilter);

            subscriber.SubscribeDataSet(subscribeDataSetArguments);

            AssertUtilities.WaitingPhase(2 * 60 * 1000);


        }

        [TestMethod]
        public void DynamicLayoutTest_Logical_Connection()
        {
            UAPubSubResolverDescriptor resolverDescriptor = UAPubSubResolverDescriptor.Publisher("opc.tcp://WBTR-W10:5000/Softing");
            var connectionDescriptor = new UAPubSubConnectionDescriptor {Name = "Connection Writer"};

            var subscriber = new EasyUASubscriber();
            subscriber.DataSetMessage += (sender, args) =>
            {
                AssertUtilities.WriteLine(args);
            };

            var subscribeDataSetArguments = new EasyUASubscribeDataSetArguments(resolverDescriptor, connectionDescriptor);

            subscriber.SubscribeDataSet(subscribeDataSetArguments);

            AssertUtilities.WaitingPhase(2 * 60 * 1000);


        }

        [TestMethod]
        public void DynamicLayoutTest_Logical_PublishedDataSetName()
        {
            UAPubSubResolverDescriptor resolverDescriptor = UAPubSubResolverDescriptor.Publisher("opc.tcp://WBTR-W10:5000/Softing");

            var subscriber = new EasyUASubscriber();
            subscriber.DataSetMessage += (sender, args) =>
            {
                AssertUtilities.WriteLine(args);
            };

            var subscribeDataSetArguments = new EasyUASubscribeDataSetArguments(resolverDescriptor, "Simple_PublishedDataSet");

            subscriber.SubscribeDataSet(subscribeDataSetArguments);

            AssertUtilities.WaitingPhase(2 * 60 * 1000);


        }
    }
}
